
/**********************************************************************

Copyright (C) 1995, 1996  E. O. Lawrence Berkeley National Laboratory.  
All Rights Reserved.  Permission to copy and modify this software and 
its documentation (if any) is hereby granted, provided that this notice
is retained thereon and on all copies.  

This software is provided as a professional academic contribution for 
joint exchange.   Thus it is experimental and scientific in nature, 
undergoing development, and is provided "as is" with no warranties of 
any kind whatsoever, no support, promise of updates or printed 
documentation.

This work is supported by the U. S. Department of Energy under contract 
number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
University of California.


	Author: Wes Bethel
		Lawrence Berkeley National Laboratory
                Berkeley, California 

  "this software is 100% hand-crafted by a human being in the USA"

***********************************************************************/

#include <stdarg.h>
#include <stdio.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "fastrak.h"

/* A very simple program to illustrate the basics of reading a Fastrak */
/* Opens a serial port, sets up a baud rate of 9600, and attempts to */
/* read positions form the Fastrak. */

/* This is a very simple program and there is no error recovery for */
/* dropped packets and the like. It is therefore best if the fastrak */
/* is power cycled before running this program. */

/* Note that this program **WILL NOT** work with Division's */
/* DRS232 device driver */

/* The fastrak **MUST** be set to this buad rate */

#define FASTRAK_BAUD B19200
#define MOUSE_BAUD B9600

#if 0
#define FASTRAK_BAUD B9600
#define MOUSE_BAUD B9600
#endif
/* The program will expect to find the fastrak on this port */
#if defined _HPUX
#define FASTRAK_DEVICE "/dev/tty00"
#endif
#if defined _SYSV42
#define FASTRAK_DEVICE "/dev/tty01"
#endif
#if defined _IRIX5
#define FASTRAK_DEVICE "/dev/ttyd2"
#define MOUSE_DEVICE "/dev/ttyd3"
#endif

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define M_PI 3.14159
#define DEGREES_TO_RADIANS(a) ((a)*M_PI/180.)
/* A static global flag which is set if byte flipping */
/* will be required on certain pieces of data */
static int needsFlipping = 0;
static int data_packet_size = 0;
static int static_nstations;

/* Local function declarations */
void writePacket (int fd, char *buffer);
int readPacket (int fd, char *buffer, int size);
static void bigEndian (void);
static void byteFlip (register int *data, int count);
static void checkPacket (char * buffer);
static void convertData (char * buffer, float * answer);

/* new stuff */
static int fastrakFD= -1;
static int mouseFD = -1;
static struct termios fastrakTermio;	/* For serial port configuration */
static struct termios mouseTermio;
static int non_canonical_min_char_xfer_size = 29; /* see termio(7), wes */

int open_fastrak(void)
{
    int fd;				/* Our file descriptor */

    /* See if our machine needs to do any byte flipping */
    /* A flag will be set if it does */
    bigEndian ();

    /* Attempt to open the port */
    /* On SYSV42 machines it may be necessary to short */
    /* pins 1 and 4 of the computer end together before */
    /* the open will work. The symptoms are that the open */
    /* hangs but is interruptable by Control-C etc. */
    printf ("Attempting to open the serial port\n");
    if ((fastrakFD = open (FASTRAK_DEVICE, O_RDWR)) < 0)
    {
	printf ("Failed to open the port with status %d : %s\n", errno, strerror (errno));
	return(0);
    }
    printf ("Opened the serial port OK\n");

    /* So we've got the device opened! */
    /* Now we have to configure the serial port so that it can */
    /* talk to the fastrak. This means setting up the port parameters */
    /* to match the fastrak settings. */
    /* They both have to speak the same language to each other :-) */
    printf ("Setting up configuration information\n");
    
    /* Set up the baud rate, bits per byte and some other voodoo. */
    /* Full details can be found in the man pages for termios etc. */
    fastrakTermio.c_cflag = FASTRAK_BAUD | CREAD | CLOCAL | CS8;

    /* These are all irrelevant in this application */
    fastrakTermio.c_iflag     = 0;
    fastrakTermio.c_oflag     = 0;
    fastrakTermio.c_lflag     = 0;

    /* In a more advanced system, these can be used to control timeouts. */
    /* Timeouts will be necessary for some degree of error recovery. */

    /* the 29 is not necessarily correct...  */
    /* Size of an EULER packet */
    fastrakTermio.c_cc[VMIN]  = non_canonical_min_char_xfer_size;
    fastrakTermio.c_cc[VTIME] = 1;

    /* We've set up the structure, so know we tell the serial port */
    /* the configuration we've just set up. We make the serial port */
    /* flush its internal buffers as we do it! */
    printf ("Sending configuration to serial port\n");
    if (tcsetattr (fastrakFD, TCSAFLUSH, &fastrakTermio))
    {
	printf ("Failed to send configuration info\n");
	printf ("Return status was %d : %s\n", errno, strerror(errno));
	exit (-1);
    }
    printf ("Sent configuration OK\n");

    tcflush(fastrakFD,TCIOFLUSH);
    
    return(1);
}

int open_3dmouse(void)
{
    int fd;				/* Our file descriptor */
    
    printf ("Attempting to open the mouse port\n");
    if ((mouseFD = open (MOUSE_DEVICE, O_RDWR)) < 0)
    {
	printf ("Failed to open the port with status %d : %s\n", errno, strerror (errno));
	return(0);
    }
    printf ("Opened the mouse port OK\n");

    /* So we've got the device opened! */
    /* Now we have to configure the serial port so that it can */
    /* talk to the fastrak. This means setting up the port parameters */
    /* to match the fastrak settings. */
    /* They both have to speak the same language to each other :-) */
    
    printf ("Setting up configuration information\n");
    
    /* Set up the baud rate, bits per byte and some other voodoo. */
    /* Full details can be found in the man pages for termios etc. */
    mouseTermio.c_cflag = MOUSE_BAUD | CREAD | CLOCAL | CS8;

    /* These are all irrelevant in this application */
    mouseTermio.c_iflag     = 0;
    mouseTermio.c_oflag     = 0;
    mouseTermio.c_lflag     = 0;

    /* In a more advanced system, these can be used to control timeouts. */
    /* Timeouts will be necessary for some degree of error recovery. */

    mouseTermio.c_cc[VMIN]  = 4; /* size of a mouse packet */
    mouseTermio.c_cc[VTIME] = 0;

    /* We've set up the structure, so know we tell the serial port */
    /* the configuration we've just set up. We make the serial port */
    /* flush its internal buffers as we do it! */
/*    printf ("Sending configuration to serial port\n"); */
    
    if (tcsetattr (mouseFD, TCSAFLUSH, &mouseTermio))
    {
	printf ("Failed to send configuration info\n");
	printf ("Return status was %d : %s\n", errno, strerror(errno));
	exit (-1);
    }
    printf ("Sent configuration OK\n");

    return(1);
}

int
close_fastrak(void)
{
    if (close(fastrakFD) == 0)
	return(1);
    else
    {
	perror("fastrak close fails");
	return(0);
    }
}


int
reset_fastrak(void)
{
    char buffer[2];
    char statusrec[64];
    int status,i;
    
    buffer[0] = 0x19;
    buffer[1] = 0;
    writePacket (fastrakFD, buffer);
    fprintf(stderr," sleeping....");
    fflush(stderr);
    sleep(10);			/* wait 10 seconds for reset to happen.*/
    fprintf(stderr,"\n");


    /* to see if the system's awake, request a system status record */

    for (i=0;i<4;i++)
    {
	buffer[0] = 'S';
	writePacket(fastrakFD, buffer);
	memset(statusrec,0,64);
	status = readPacket(fastrakFD,statusrec,55);
	if (status == 55)
	    break;
    }
    if (i==4)
    {
	fprintf(stderr,"\tFailed to get any response from the fastrak.  Check cabling, etc.\n");
	return(0);
    }

    return(1);
}

int
configure_fastrak_stations(station_info *sa)
{
    /* in this routine, we figure out how many stations
     we have, and set the output format for each of them. */
    
    char buffer[128];
    int i,status,nstations=0;

    /* send the "active station" command to get a listing of the
      connected stations */
    
    memset(buffer,0,128);
    buffer[0] = 'l';
    buffer[1] = '1';
    buffer[2] = ',';
    buffer[3] = '\r';

    writePacket(fastrakFD,buffer);
    memset(buffer,0,128);
    status = readPacket(fastrakFD,buffer,9);
    if (buffer[2] == 'E')	/* an error record */
    {
	fprintf(stderr,"\tError: received an error record instead of a station status record. aborting..\n");
	exit(-1);
    }
    if (status == 9)
    {
	for (i=0;i<MAX_STATIONS;i++)
	{
	    if (buffer[i+3] == '1')
	    {
		sa[nstations].station_num = i+1;
		sa[nstations].flags = THIS_CONFIG;
		nstations++;
	    }
	}
    }

    /* all stations will use binary data transfer mode. */
    memset(buffer,0,128);
    buffer[0] = 'f';
    writePacket(fastrakFD,buffer);

    /* all stations will use on-demand data xfer mode. */
    buffer[0] = 'c';
    writePacket(fastrakFD,buffer);

    /* grab another system status record to verify that these changes
     took place.*/
    for (i=0;i<4;i++)
    {
	buffer[0] = 'S';
	writePacket(fastrakFD, buffer);
	memset(buffer,0,64);
	status = readPacket(fastrakFD,buffer,55);
	if (status == 55)
	    break;
    }
    /* check for binary data xfer, non-continuous data xfer. */
    if(!((buffer[3] & 0x1) && (buffer[3] & 0x2)))
    {
	fprintf(stderr,"\tError: status record doesn't reflect non-continuous data and/or binary data xfer.\n");
    }

    /* set the output format for each of the active stations */
    for (i=0;i<nstations;i++)
    {
	char loc_buffer[32];
	sprintf(loc_buffer,"O%d,%s\r",sa[i].station_num,output_format);
	writePacket(fastrakFD,loc_buffer);
    }

    data_packet_size = nstations * BYTES_PER_STATION;
    static_nstations = nstations;
    
    return(nstations);
}

int
request_fastrak_data(void)
{
    char buffer[256];
    int status,i,index;

    memset(buffer,0,256);
    buffer[0] = 'P';		/* a request-data command */
    writePacket(fastrakFD,buffer);
    
}

int
input_fastrak_data(station_info *sa)
{
    static fd_set zero_set,ft_set;
    char buffer[256];
    int i,index;

    static struct timeval longtimeval = { 1, 0 };
    int status;

    FD_ZERO( &zero_set );
    FD_SET( fastrakFD, &ft_set);

    /* return is 0 if no data, positive if data, negative if error */

    status = select(fastrakFD+1, &ft_set, NULL, NULL, &longtimeval);

    if (status)
    {
	memset(buffer,0,256);
	if (readPacket(fastrakFD,buffer,data_packet_size) != data_packet_size)
	    return(0);
	
	for (i=0;i<static_nstations;i++)
	{
	    index = i * BYTES_PER_STATION;
	    decode_raw_data(buffer,index,sa);
	}
    }
    return(status);
}

int
get_fastrak_data(station_info *sa)
{
    char buffer[256];
    int status,i,index;

    memset(buffer,0,256);
    buffer[0] = 'P';		/* a request-data command */
    writePacket(fastrakFD,buffer);
    
    memset(buffer,0,256);
    if (readPacket(fastrakFD,buffer,data_packet_size) != data_packet_size)
	return(0);
    
    for (i=0;i<static_nstations;i++)
    {
	index = i * BYTES_PER_STATION;
	decode_raw_data(buffer,index,sa);
    }

    return(1);
}


#if 0
  /* Once we've sent the configuration to the serial port */
  /* we should be able to speak with the fastrak! */
  /* Sending the correct bytes to the fastrak is how we */
  /* send commands to it. It may reply with some response */
  /* bytes. Full details are in the Fastrak manual */

  /* A more sophisticated program would do some work to discover */
  /* which sensors are connected and operational. This simple */
  /* program assumes that there is an operational sensor */
  /* connected to socket 1 on the fastrak. You could use */
  /* "testipu" to verify that this is working */

  /* We are going to grab EULER type packets from the fastrak */
  /* It doesn't know this yet, and so we tell it that this */
  /* type of packet is required from sensor 1 */
  printf ("Setting up for EULER type packets\n");
  writePacket (fastrakFD, "fu");		/* Binary output, metric */
  writePacket (fastrakFD, "O1,2,4,1\r");
  printf ("Set up for EULER packet OK!\n");

  /* We've completed the set up process, and so now we can enter */
  /* an infinite loop getting posistions from the sensor and */
  /* printing them. There is no checking for dropped bytes here */
  /* so missed data may corrupt the co-ordinates etc. */
  while (1)
  {
    char buffer [40];		/* A buffer with enough room for 10 floats */
    float data  [10];

    /* Request a packet from the fastrak */
    /* This is achieved by sending a "P" */

#ifdef VERBOSE
    printf ("Sending data request\n");
#endif
    writePacket (fd, "P");
    tcdrain (fd);
#ifdef VERBOSE
    printf ("Sent data request\n");
#endif

    /* And read the data as it comes in. */
    /* An EULER packet will be 29 bytes in size, and have 6 floats in it */
#ifdef VERBOSE
    printf ("Waiting for response packet\n");
#endif
    readPacket (fd, buffer, 29);
#ifdef VERBOSE
    printf ("Got response packet\n");
#endif

    /* We can now validate the packet for integrity */
#ifdef VERBOSE
    printf ("Checking packet\n");
#endif
    checkPacket (buffer);
#ifdef VERBOSE
    printf ("Checked packet OK\n");
#endif

    /* Convert the data */
#ifdef VERBOSE
    printf ("Converting data\n");
#endif
    convertData (buffer, data);
#ifdef VERBOSE
    printf ("Converted data\n");
#endif

    /* Phew! Print the numbers out */
#if 0
    printf ("FASTRAK READING GAVE: %3.2f %3.2f %3.2f %3.2f %3.2f %3.2f\n",
                                   data[0],	/* First 3 are X, Y, Z */
                                   data[1],
                                   data[2],
                                   data[3],	/* Next 3 are Euler angles */
                                   data[4],
                                   data[5]);
#endif
    printf ("FASTRAK READING GAVE: %3.2f %3.2f %3.2f %3.2f %3.2f %3.2f\r",
                                   data[0],	/* First 3 are X, Y, Z */
                                   data[1],
                                   data[2],
                                   data[3],	/* Next 3 are Euler angles */
                                   data[4],
                                   data[5]);
    fflush(stdout);
  }
}
#endif

/* Write a packet with error validation */
void
writePacket (int fd, char *buffer)
{
    struct timeval tv1,tv2;
    int length = strlen (buffer);
    int result;

#if 0
    if (gettimeofday(&tv1) != 0)
	fprintf(stderr," couldn't get time value \n");
#endif
    
    /* clean the line */
   tcflush(fastrakFD,TCIOFLUSH);    
  
    if ((result = write (fd, buffer, length)) != length)
    {
	printf ("Failed to write packet of %d bytes with status %d\n", length, result);
	if (result < 0)
	{
	    printf ("UNIX system error was %d : %s\n", errno, strerror (errno));
	}
	exit (-1);
    }
    tcdrain(fd); 

#if 0
    if (gettimeofday(&tv2) != 0)
	fprintf(stderr," couldn't get time value \n");
    fprintf(stderr,"\nelapsed time to transfer %d chars: %d\t\t%d\n",length,tv2.tv_sec-tv1.tv_sec,tv2.tv_usec-tv1.tv_usec);
#endif
}	


/* Read a packet with error validation */
int
readPacket (int fd, char *buffer, int size)
{
    int result;

    /* check to see if the number of characters we're requesting this time
       is the same that we requested last time.  if not, then reset the
       terminal line characteristics to ask for that many characters at
       a whack. */

    if (size != non_canonical_min_char_xfer_size)
    {
/*      fastrakTermio.c_cc[VMIN]  = non_canonical_min_char_xfer_size; */
	fastrakTermio.c_cc[VMIN]  = size;
	fastrakTermio.c_cc[VTIME] = 2;
	non_canonical_min_char_xfer_size = size;
	if (tcsetattr (fastrakFD, TCSANOW, &fastrakTermio))
	{
	    printf ("Failed to reconfigure input line\n");
	    printf ("Return status was %d : %s\n", errno, strerror(errno));
	    exit (-1);
	}
    }

    if ((result = read (fd, buffer, size)) != size)
    {
#ifdef DEBUG
	printf ("Failed to read packet of %d bytes with status %d\n", size, result);
#endif
	if (result < 0)
	{
	    printf ("UNIX system error was %d : %s\n", errno, strerror (errno));
	}
    }
    return(result);
}


/* Are we bigEndian? */
/* Some data may need byte flipping if we are */
static void bigEndian (void)
{
  int a = 0;

  *((char *)&a) = 1;

  if (a == 1)
    needsFlipping = 0;
  else
    needsFlipping = 1;
}

/* Flip the bytes of an int. NB Does this in SITU */
static void byteFlip (register int *data, int count)
{
  int i;

  if (needsFlipping)
  {
    for (i=0;i<count;i++)
    {
      data[i] =  ((data[i] >> 24) & 0xFF) |
		 ((data[i] >> 8)  & 0xFF00) |
		 ((data[i] << 8)  & 0xFF0000) |
		 ((data[i] << 24));
    }
  }
}

#if 0
static void checkPacket (char * buffer)
{
  if (buffer[0] != '0')
  {
    printf ("First byte of packet is not a 0\n");
    exit (-1);
  }
  if (buffer[1] != '1')
  {
    printf ("Station number is not 1\n");
    exit (-1);
  }
  if (buffer[27] != 0xD)
  {
    printf ("Byte 27 of packet is not a 0xD\n");
    exit (-1);
  }
  if (buffer[28] != 0xA)
  {
    printf ("Byte 28 of packet is not a 0xA\n");
    exit (-1);
  }
}

static void convertData (char * buffer, float * answer)
{
  /* The first three bytes are useless header information */
  memcpy (answer, buffer + 3,  6 * sizeof(float));

  /* Byte flip the data if required */
  byteFlip ((int *)answer, 6);
}
#endif


void bflip (float *addr,int  count)
{
    int i;
    char *pointer;
    
    char b, b1, b2, b3;
    for (i = 0; i< count; i++)
    {
        pointer = (char *)addr;
        b = pointer[0];
        b1 = pointer[1];
        b2 = pointer[2];
        b3 = pointer[3];
        pointer[3] = b;
        pointer[2] = b1;
        pointer[1] = b2;
        pointer[0] = b3;
        addr++;
    }
}

int
decode_raw_data(char *buffer,
		int index,
		station_info *sa)
{
    int j,sn;
    char cwork[2];
/*    float locPos[6]; */
    float fbuf[3];
    int loc_index=0;
    
    if (buffer[index] != '0')	/* check for record type zero */
    {
	fprintf(stderr,"\terror: expected record type '0', got type '%c'.\n",buffer[index]);
	return(0);
    }

    cwork[0] = buffer[index+1]; /* the station number */
    sscanf(cwork,"%d",&j);
    sn = j;
    sa[sn-1].station_num = j;

#if 0
    memcpy (locPos, buffer+index+3, sizeof (float) *6);
    bflip (locPos, 6);
#endif
    /* we make lots of assumptions about how things are ordered as they
     come in.  the general rule of thumb is that things will be coming
     in in the order of the opcodes specified in the 3space manual.  e.g.,
     the opcode for x-direction cosines is 5, and the opcode for the
     y-direction cosines is 6.  thus, the x direction cosines will appear
     in the input stream before the y-direction cosines. */

    loc_index = index;
    if (THIS_CONFIG & POSITION_FLAG)
    {
	memcpy(fbuf,buffer+loc_index+3,sizeof(float)*3);
	bflip(fbuf,3);
	loc_index += 3*sizeof(float);
	
    /*
     * get the positions/orientations out of the record, relabelling the
     * axis as we do so, also add any local position and orientation
     * adjustments at this time.
     */

	if (sn == 1)		/* assume it's the hand */
	{
	    sa[sn-1].pos[1]= fbuf[0];
	    sa[sn-1].pos[2] = fbuf[1];
	    sa[sn-1].pos[0] = -fbuf[2];
	}
	else
	{
	    sa[sn-1].pos[0]= -fbuf[0];
	    sa[sn-1].pos[2] = fbuf[1];
	    sa[sn-1].pos[1] = -fbuf[2];
	}

    }

    if (THIS_CONFIG & RELATIVE_POSITION_FLAG)
    {
	memcpy(fbuf,buffer+loc_index+3,sizeof(float)*3);
	bflip(fbuf,3);
	loc_index += 3*sizeof(float);
	
    /*
     * get the positions/orientations out of the record, relabelling the
     * axis as we do so, also add any local position and orientation
     * adjustments at this time.
     */

	if (sn == 1)		/* assume it's the hand held mouse */
	{
#if 0
	    sa[sn-1].delta_pos[1]= fbuf[0];
	    sa[sn-1].delta_pos[2] = fbuf[1];
	    sa[sn-1].delta_pos[0] = -fbuf[2];
#endif
	    sa[sn-1].delta_pos[1]= fbuf[1];
	    sa[sn-1].delta_pos[2] = fbuf[2];
	    sa[sn-1].delta_pos[0] = fbuf[0];
	}
	else
	{
#if 0
	    sa[sn-1].delta_pos[0]= -fbuf[0];
	    sa[sn-1].delta_pos[2] = fbuf[1];
	    sa[sn-1].delta_pos[1] = -fbuf[2];
#endif
	    sa[sn-1].delta_pos[0] = fbuf[0];
	    sa[sn-1].delta_pos[1] = fbuf[1];
	    sa[sn-1].delta_pos[2] = fbuf[2];
	}
    }

    if (THIS_CONFIG & ROTATION_FLAG)
    {
	memcpy(fbuf,buffer+loc_index+3,sizeof(float)*3);
	bflip(fbuf,3);
	loc_index += 3*sizeof(float);

	if (sn == 1)		/* assume it's the hand */
	{
#if 0
	    sa[sn-1].rot[0]=( -fbuf[0]);
	    sa[sn-1].rot[2]=(  fbuf[1]);
	    sa[sn-1].rot[1]=( -fbuf[2]);
#endif

	    sa[sn-1].rot[0]=( fbuf[0]);
	    sa[sn-1].rot[1]=( fbuf[1]);
	    sa[sn-1].rot[2]=( fbuf[2]);

	}
	else
	{
#if 0
	    sa[sn-1].rot[1]=( -fbuf[0]);
	    sa[sn-1].rot[2]=(  fbuf[1]);
	    sa[sn-1].rot[0]=( -fbuf[2]);
#endif
	    sa[sn-1].rot[0]=( fbuf[0]);
	    sa[sn-1].rot[1]=( fbuf[1]);
	    sa[sn-1].rot[2]=( fbuf[2]);
	}
    }
    
    if (THIS_CONFIG & XDIRCOS_FLAG)
    {
	memcpy(fbuf,buffer+loc_index+3,sizeof(float)*3);
	bflip(fbuf,3);
	loc_index += 3*sizeof(float);

	sa[sn-1].xdircos[0]=( fbuf[0]);
	sa[sn-1].xdircos[1]=(  fbuf[1]);
	sa[sn-1].xdircos[2]=( fbuf[2]);
    }
    
    if (THIS_CONFIG & YDIRCOS_FLAG)
    {
	memcpy(fbuf,buffer+loc_index+3,sizeof(float)*3);
	bflip(fbuf,3);
	loc_index += 3*sizeof(float);

	sa[sn-1].ydircos[0]=( fbuf[0]);
	sa[sn-1].ydircos[1]=(  fbuf[1]);
	sa[sn-1].ydircos[2]=( fbuf[2]);
    }
    
    if (THIS_CONFIG & ZDIRCOS_FLAG)
    {
	memcpy(fbuf,buffer+loc_index+3,sizeof(float)*3);
	bflip(fbuf,3);
	loc_index += 3*sizeof(float);

	sa[sn-1].zdircos[0]=( fbuf[0]);
	sa[sn-1].zdircos[1]=(  fbuf[1]);
	sa[sn-1].zdircos[2]=( fbuf[2]);
    }
}

int
filter_enable(void)
{
    char buffer[132];
#if 0
    strcpy(buffer,"2 x,00.200, 00.200, 00.800, 00.800\r");
    write (fd, buffer,strlen(buffer)); /* enable filtering */
#endif

    /* --- first deal with the positional filter --- */
    /* read the current filter settings.. */
    writePacket(fastrakFD,"x\r");
    memset((char *)buffer,0,132);
    readPacket(fastrakFD,buffer,33);

    /* set new values.. */
    writePacket(fastrakFD,"x 0.1,0.03,0.95,0.95\r");

    /* read the new filter values. */
    writePacket(fastrakFD,"x\r");
    memset((char *)buffer,0,132);
    readPacket(fastrakFD,buffer,33);

    /* --- then deal with the attitude or rotational filter --- */
    /* read the current filter settings.. */
    writePacket(fastrakFD,"v\r");
    memset((char *)buffer,0,132);
    readPacket(fastrakFD,buffer,33);

    /* set new values.. */
    writePacket(fastrakFD,"v 0.02,0.03,0.95,0.95\r");

    /* read the new filter values. */
    writePacket(fastrakFD,"v\r");
    memset((char *)buffer,0,132);
    readPacket(fastrakFD,buffer,33);

}

int
reset_alignment_frame(int station,float *origin, float *xaxis, float *yaxis)
{
    char buffer[256];

    memset(buffer,0,256);
    sprintf(buffer,"A%d,%9.8f,%9.8f,%9.8f,%9.8f,%9.8f,%9.8f,%9.8f,%9.8f,%9.8f\r",station,origin[0],origin[1],origin[2],xaxis[0],xaxis[1],xaxis[2],yaxis[0],yaxis[1],yaxis[2]);


    writePacket(fastrakFD,buffer);
#if 0
    memset(buffer,0,256);
    sprintf(buffer,"A1\r");
    writePacket(fastrakFD,buffer);

    readPacket(fastrakFD,buffer,68);
#endif
}

int
get_mouse_data(int *bmask)
{
    char buffer[4],buttons[4];
    int failed, missing,i,j;

    readMousePacket(mouseFD,buffer,4);
    
    tcflush(mouseFD,TCIFLUSH);

    /* Now we should have the most recent four bytes in the buffer */
    /* We check to make sure that they are all OK */
    /* This makes sure that they are in the right order */
    /* In this program there is no sophisticated recovery in otder that */
    /* we get back to the right order. We rely on the flush */
    
    failed = 0;
    for (i=0;i<4;i++)
    {
	missing = 1;
	for (j=0;j<4;j++)
	{
	    if (((buffer[j] & 0xC0) >> 6) == i)
	    {
		buttons[i] = buffer[j];
		missing = 0;
	    }
	}
	if (missing)
	    failed = 1;
    }

    missing = 1;
    if (!(buttons[0] & 0x1))
    {
/*	fprintf (stderr,"MIDDLE\n"); fflush(stderr); */
	bmask[1] = 1;
	missing = 0;
    }
    if (!(buttons[0] & 0x2))
    {
/*	fprintf (stderr,"TOP\n"); fflush(stderr); */
	missing = 0;
	bmask[3] = 1;
    }
    if (!(buttons[0] & 0x4))
    {
/*	fprintf (stderr,"RIGHT\n"); fflush(stderr); */
	bmask[2] = 1;
	missing = 0;
    }
    if (!(buttons[0] & 0x8))
    {
/*	fprintf (stderr,"LEFT\n"); fflush(stderr); */
	bmask[0] = 1;
	missing = 0;
    }
    if (!(buttons[0] & 0x10))
    {
/*	fprintf (stderr,"BOTTOM\n"); fflush(stderr); */
	bmask[4] = 1;
	missing = 0;
    }
#if 0
    if (missing)
	fprintf ("NONE\n");
#endif
    if (missing)
	return(0);
    else
	return(1);

}

int readMousePacket (int fd, char *buffer, int size)
{
  int result;

  if ((result = read (fd, buffer, size)) != size)
  {
    printf ("Failed to read packet of %d bytes with status %d\n", size, result);
    if (result < 0)
    {
      printf ("UNIX system error was %d : %s\n", errno, strerror (errno));
    }
  }
  else
      result = result;
}

int
reset_boresight(int station_num,
		float *r)
{
    char buffer[128];
    /* check to see what boresight values are */
#if 0
    memset(buffer,0,128);
    sprintf(buffer,"G%d,\r",station_num);
    writePacket(fastrakFD,buffer);
    memset(buffer,0,128);
    readPacket(fastrakFD,buffer,26);
#endif
    /* set new boresight values.. */
    memset(buffer,0,128);
    sprintf(buffer,"G%d,%6.3f,%6.3f,%6.3f\r",station_num,-r[0],r[2],-r[1]);
    writePacket(fastrakFD,buffer);


    /* and activate them. */
    memset(buffer,0,128);
    sprintf(buffer,"B%d\r",station_num);
    writePacket(fastrakFD,buffer);


#if 0
    /* check to see what the new boresight values are */
    memset(buffer,0,128);
    sprintf(buffer,"G%d,\r",station_num);
    writePacket(fastrakFD,buffer);
    memset(buffer,0,128);
    readPacket(fastrakFD,buffer,26);
#endif
}


int
realign_stations(station_info *sa,
		 int nstations)
{
    int i;
    float dorigin[3];

    /* make a dummy origin */
    dorigin[0] = dorigin[1] = dorigin[2] = 0.; 

    /* assume that device 1 is the 3dmouse */
    reset_alignment_frame(1,mouse3d.origin,mouse3d.xaxis,mouse3d.yaxis);

    /* and that device 2, if present, is the vr4 */
    if (nstations > 1)
	reset_alignment_frame(2,vr4.origin,vr4.xaxis,vr4.yaxis);
}

int
build_composite_matrix(matrix4x4 *m,float *r)
{
    /**
      * NOTES:
      *
      * what we're going to do here is build a transformaion matrix and
      * put it into "m".  this transformation matrix is formed by the absolute
      * rotation angles present in float r[3].  the way in which r[] is
      * interpreted is r[0] = azimuth, r[1] = elevation, r[2] = roll.
      * furthermore, we're making some assumptions with regard to the
      * fastrak device.  these are:
      *
      * 1. the alignment reference frame for all stations is the power on
      *    reference frame: a right handed coordinate systems with +z
      *    pointing up.
      *
      * 2. azimuth represents angular changes about the z axis of the
      *    alignment reference frame; elevation represents changes aboiut
      *    the y axis of the alignment reference frame; roll represents
      *    angular changes about the x-axis of the alignment reference
      *    frame.
      * 3. the angular values presented in r[] are taken in the order of
      *    azimuth, elevation then roll.  the transformation matrix we
      *    construct below will be computed in this order.
      *
      * the sequence of computations below is valid only if all these
      * assumptions are TRUE.
      *
      * the routine which is called subsequent to this routine,
      * "recompute_base_rotations" will recompute delta azimuth, elevation
      * and roll values making use of the same 3 fastrak assumptions as this
      * routine.
    **/
      
    matrix4x4 azimuth, elevation, roll;
    double t,c,s;
    extern double cos(),sin();
    int i;

    /* do azimuth - according to the fastrak manual, azimuth angles
     occur about the "z" axis.*/
    identity_4x4(&azimuth);
    t = DEGREES_TO_RADIANS(r[0]);
    c = cos(t);
    s = sin(t);
    azimuth.m[0][0] = azimuth.m[1][1] = c;
    /**
      * NOTE: the following formulation for computing the z-axis rotation
      * is inconsistent with that outlined in foley/van dam and various
      * other graphics texts.
      *
      * the way that this routine was formulated was by obtaining from
      * the device it's direction cosines, and comparing those with the
      * matrix produced by this routine.
      *
      * the matrix produced here has the following relationship with the
      * direction cosines produced by the device:
      *
      * 1. row 0 of the matrix corresponds to the x direction cosine exactly
      *    (subject to floating point accuracy)
      * 2. row 1 of the matrix corresponds to the y direction cosines exactly
      *    (again, subject to floating point accuracy)
      * 3. row 2 of the matrix corresponds to the z-direction cosines, except
      *    the signs are wrong.  thus, we'll change the signs explicitly
      *    after constructing the matrix.
      *
     **/
    azimuth.m[0][1] = -s;	
    azimuth.m[1][0] = s;	

    /* do elevation */
    t = DEGREES_TO_RADIANS(r[1]); /* rotation about y axis */
    c = cos(t);
    s = sin(t);
    identity_4x4(&elevation);
    elevation.m[0][0] = elevation.m[2][2] = c;
    elevation.m[0][2] = s;
    elevation.m[2][0] = -s;

    /* do roll - according to the manual, roll occurs about the X axis. */
    identity_4x4(&roll);
    t = DEGREES_TO_RADIANS(r[2]); /* rotation about x axis */
    c = cos(t);
    s = sin(t);
    roll.m[1][1] = roll.m[2][2] = c;
    roll.m[1][2] = -s;
    roll.m[2][1] = s;

    identity_4x4(m);

    mmul_4x4(m,&azimuth);
    mmul_4x4(m,&elevation);
    mmul_4x4(m,&roll);


}

int
recompute_base_rotations(mthis,mlast,azimuth,elevation,roll)
matrix4x4 *mthis;
matrix4x4 *mlast;
float *azimuth;
float *elevation;
float *roll;
{
    /**
     * "roll" occurs about the X axis
     * "azimuth" occurs about the Z axis
     * "elevation" occurs about the Y axis
     **/
    
    extern double acos();
    int k;
    double t1,t2;
    vector3 xold,yold,zold,xnew,ynew,znew;
    vector3 w1,w2;
    matrix4x4 delta,tmp;

    transpose_4x4(mlast,&delta);
    mmul_4x4(&delta,mthis);
    
    for (k=0;k<3;k++)
    {
		    
	xnew.v[k] = delta.m[k][0];
	ynew.v[k] = delta.m[k][1];
	znew.v[k] = delta.m[k][2];
#if 0
	xold.v[k] = mlast->m[k][0];
	yold.v[k] = mlast->m[k][1];
	zold.v[k] = mlast->m[k][2];
		    
	xnew.v[k] = mthis->m[k][0];
	ynew.v[k] = mthis->m[k][1];
	znew.v[k] = mthis->m[k][2];
#endif
    }

    xold.v[0] = 1.0;
    xold.v[1] = xold.v[2] = 0.;

    yold.v[1] = 1.;
    yold.v[0] = yold.v[2] = 0.;

    zold.v[2] = 1.;
    zold.v[0] = zold.v[1] = 0.;

    /* compute amount of change about the Z axis - this is azimuth */


    w1.v[0] = xnew.v[0];
    w1.v[1] = xnew.v[1];
    w1.v[2] = 0.;		/* project into x-axis into x-y plane */
    vert3norm(&w1,&w1);		/* and normalize the projected vector */
    vec3dot(&xold,&w1,&t1);
    
    w1.v[0] = xnew.v[0];
    w1.v[1] = xnew.v[1];
    w1.v[2] = 0.;		/* project into y-axis into x-y plane */
    vert3norm(&w1,&w1);		/* and normalize the projected vector */
    vec3dot(&yold,&w1,&t2);

    t1 = MAX(t1,t2);
    if (t1 > 1.)
	t1 = 1.;
    t1 = acos(t1);
    vec3dot(&xnew,&yold,&t2);
    t1 *= ((t2 > 0.) ? 1. : -1.);
    *azimuth = (float)t1;

    /* compute the amount of change about the Y axis. this is elevation */
    
    w1.v[0] = xnew.v[0];
    w1.v[1] = 0.;	/* project into x-axis into x-z plane */
    w1.v[2] = xnew.v[2];
    vert3norm(&w1,&w1);		/* and normalize the projected vector */
    vec3dot(&xold,&w1,&t1);

    w1.v[0] = znew.v[0];
    w1.v[1] = 0;		/* project z-axis into the x-z plane */
    w1.v[2] = znew.v[2];
    vert3norm(&w1,&w1);		/* and noramalize the projected vector */
    vec3dot(&zold,&w1,&t2);
    
    t1 = MAX(t1,t2);
    if (t1 > 1.)
	t1 = 1.;
    t1 = acos(t1);
    vec3dot(&xnew,&zold,&t2);
    t1 *= ((t2 > 0.) ? 1. : -1.);
    *elevation = (float)t1;

    /* compute the amount of change about the X axis.  this is roll. */
    w1.v[0] = 0.;		/* project the y-axis into the y-z plane */
    w1.v[1] = ynew.v[1];
    w1.v[2] = ynew.v[2];
    vert3norm(&w1,&w1);		/* and normalize the projected vector */
    vec3dot(&yold,&w1,&t1);

    w1.v[0] = 0.;		/* project the z-axis into the y-z plane */
    w1.v[1] = znew.v[1];
    w1.v[2] = znew.v[2];
    vert3norm(&w1,&w1);		/* and normalize the projected vector */
    vec3dot(&zold,&w1);

    t1 = MAX(t1,t2);
    if (t1 > 1.)
	t1 = 1.;
    t1 = acos(t1);
    vec3dot(&znew,&yold,&t2);
    t1 *= ((t2 > 0.) ? 1. : -1.);
    *roll = (float)t1;
}


